from __future__ import annotations

from typing import Dict, Generator, Iterable, TYPE_CHECKING, Optional
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.utils.markdown2 import markdown

from bs4 import BeautifulSoup
from bs4.element import PageElement, Tag
import re
from pyvisflow.core.components import Component
from pyvisflow.core.components.markdownValue import MarkdownValue

if TYPE_CHECKING:
    from pyvisflow.core.props import TypePropInfo

RE_CONTROLER = re.compile('{{.+?}}')


class MarkdownLayout():
    def __init__(self,
                 md: str,
                 props_mapping: Optional[Dict[str, TypePropInfo]] = None
                 ) -> None:
        self._md = md
        self._props_mapping = props_mapping or {}
        self._component_mapping: Dict[str, MarkdownValue] = {}

    def __update_prop2cp(self, cp_name: str, prop: TypePropInfo):
        assert cp_name in self._component_mapping, f'not found name[{cp_name}] in markdown'
        # self._component_mapping[cp_name].update_attrs(**{
        #     'id': prop.reactive._id, 'path': prop.get_trigger_path(), 'logic': prop.get_post_change_logic('return change')
        # })
        self._component_mapping[cp_name].set_value(prop)

    def parse_markdown(self, ):
        def _gen_from_string(s: str) -> Generator[Component, None, None]:
            mats = list(RE_CONTROLER.finditer(s))

            start = 0
            for m in mats:
                if m.start() > start:
                    yield Component(tag=s[start:m.start()],
                                    type=TComponentType.htmlRawString)
                    # yield Controler(tag=s[start:m.start()], type=ControlerTypeEnum.htmlRawString, responsive=False, children=[])

                cp_name = m.group().replace('{{', '').replace('}}', '')
                self._component_mapping[cp_name] = MarkdownValue()

                if cp_name in self._props_mapping:
                    prop = self._props_mapping[cp_name]
                    self.__update_prop2cp(cp_name, prop)

                yield self._component_mapping[cp_name]

                start = m.end()

            if start == 0:
                yield Component(tag=s, type=TComponentType.htmlRawString)

        def _gen_nodes(tags: Iterable[PageElement] or str, ):
            for t in tags:
                if type(t) is Tag:
                    cp = Component(tag=t.name, type=TComponentType.htmlRaw)
                    cp._children = list(_gen_nodes(t.children))
                    yield cp
                else:
                    if t == '\n':
                        continue
                    yield from _gen_from_string(t)

        html = markdown(
            self._md, extras=['fenced-code-blocks', 'task_list', 'code-color'])
        soup = BeautifulSoup(html, 'html.parser')
        return _gen_nodes(soup.children)

    def set(self, **prop_mapping: TypePropInfo):
        for name, prop in prop_mapping.items():
            self.__update_prop2cp(name, prop)



from typing import Dict, List
from pyvisflow.core.props import StrTypePropInfo, SubscriptableTypePropInfo
from pyvisflow.core.props.nameProp import NamePropInfo
from pyvisflow.core.props.propValuator import PropValuator
from pyvisflow.models.TComponent import TComponentType
from .components import Component


class Plotly(Component):
    def __init__(self) -> None:
        super().__init__('plotly', TComponentType.builtIn)
        self._data: List[Dict] = []
        self._layout = {}

    @property
    def clickInfo(self):
        p = self.get_prop('clickInfo')
        return SubscriptableTypePropInfo[str, StrTypePropInfo](p)

    @property
    def hoverInfo(self):
        p = self.get_prop('hoverInfo')
        # sub = SubscriptableTypePropInfo[str, SubscriptableTypePropInfo](p)
        return HoverInfoPropInfo(p)

    def from_fig(self, fig):
        fig_data = fig.to_dict()
        if 'data' in fig_data:
            for data in fig_data['data']:
                self.add_data(data)

        if 'layout' in fig_data:
            self.update_layout(fig_data['layout'])
        return self

    def add_data(self, data: Dict):
        self._data.append(data)
        return self

    def update_data(self, idx: int, data: Dict):
        self._data[idx].update(data)
        return self

    def update_layout(self, layout: Dict):
        self._layout.update(layout)
        return self

    def _ex_get_react_data(self):
        data = super()._ex_get_react_data()
        data.update({
            'data': self._data,
            'layout': self._layout,
            'config': {},
            'clickInfo': {},
            'hoverInfo': {
                'infos': {
                    'x': None,
                    'y': None
                }
            },
        })

        return data


class HoverInfoPropInfo(SubscriptableTypePropInfo):
    def __init__(self, parent) -> None:
        super().__init__(parent)

    @property
    def x(self):
        n = NamePropInfo(self.valuator, self, 'infos.x')
        return StrTypePropInfo(n)

    @property
    def y(self):
        n = NamePropInfo(self.valuator, self, 'infos.y')
        return StrTypePropInfo(n)


from __future__ import annotations

from abc import abstractmethod
from typing import Callable, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, Union

from pyvisflow.utils.helper import value2code

from .absProp import AbsPropInfo

from pyvisflow.core.props.propValuator import DataFilterValuator


class DataFilterPropInfo(AbsPropInfo):
    def __init__(self, var: str) -> None:
        super().__init__(DataFilterValuator(var), None)
        self.var = var

    def _ex_gen_expr(self):
        return self.var



from typing import Dict, List, Optional
from pydantic import BaseModel
from enum import Enum


class TComponentType(str, Enum):
    htmlRaw = 'html-raw'
    htmlRawString = 'html-raw-string'
    builtIn = 'built-in'


class TComponent(BaseModel):
    id: str
    tag: str
    type: TComponentType
    attrs: Optional[Dict] = {}
    styles: Optional[Dict] = {}
    children: Optional[List['TComponent']] = []



from typing import Any, Dict, Iterable, List, Tuple, Union

from pydantic import BaseModel
from pyvisflow.core.props import SubscriptableTypePropInfo, StrTypePropInfo, NumberTypePropInfo
from pyvisflow.core.props.nameProp import NamePropInfo

from .auto_create._select import _Select


class TSelectOption(BaseModel):
    label: str
    value: int


class Select(_Select):
    def __init__(self, options: Iterable[str], multiple=False) -> None:
        super().__init__()
        options = list(options)

        self._options = options

        self.multiple = multiple

    @property
    def options(self):
        '''
        '''

        p = self.get_prop('options')
        return SubscriptableTypePropInfo[str,StrTypePropInfo](p)

    @options.setter
    def options(self, value: Union[SubscriptableTypePropInfo[str,StrTypePropInfo], bool]):
        '''
        '''
        self.set_prop('options', value)

    @property
    def currentLabels(self):
        p = self.get_prop('currentLabels')
        return SubscriptableTypePropInfo[int, StrTypePropInfo](p)

    def _ex_get_react_data(self):
        data = super()._ex_get_react_data()
        data.update({'options': self._options, 'currentLabels': []})
        return data



from typing import Union,Dict,List
from pyvisflow.core.components.components import Component
from pyvisflow.core.props import TypePropInfo,StrTypePropInfo,SubscriptableTypePropInfo, NumberTypePropInfo, BoolTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.core.props import *


class _InputNumber(Component):
    '''
    Input Number 数字输入框
仅允许输入标准的数字值，可定义范围
    '''

    def __init__(self) -> None:
        super().__init__('input-number', TComponentType.builtIn)

    def _ex_get_react_data(self):
        return super()._ex_get_react_data()


    @property
    def value(self):
        '''
        可选值：— \n
        选中项绑定值
        '''

        
        p = self.get_prop('value')
        return NumberTypePropInfo(p)

        


    @value.setter
    def value(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：— \n
        选中项绑定值
        '''
        self.set_prop('value', value)

    @property
    def min(self):
        '''
        可选值：— \n
        设置计数器允许的最小值
        '''

        
        p = self.get_prop('min')
        return NumberTypePropInfo(p)

        


    @min.setter
    def min(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：— \n
        设置计数器允许的最小值
        '''
        self.set_prop('min', value)

    @property
    def max(self):
        '''
        可选值：— \n
        设置计数器允许的最大值
        '''

        
        p = self.get_prop('max')
        return NumberTypePropInfo(p)

        


    @max.setter
    def max(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：— \n
        设置计数器允许的最大值
        '''
        self.set_prop('max', value)

    @property
    def step(self):
        '''
        可选值：— \n
        计数器步长
        '''

        
        p = self.get_prop('step')
        return NumberTypePropInfo(p)

        


    @step.setter
    def step(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：— \n
        计数器步长
        '''
        self.set_prop('step', value)

    @property
    def step_strictly(self):
        '''
        可选值：— \n
        是否只能输入 step 的倍数
        '''

        
        p = self.get_prop('step-strictly')
        return BoolTypePropInfo(p)

        


    @step_strictly.setter
    def step_strictly(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：— \n
        是否只能输入 step 的倍数
        '''
        self.set_prop('step-strictly', value)

    @property
    def precision(self):
        '''
        可选值：— \n
        数值精度
        '''

        
        p = self.get_prop('precision')
        return NumberTypePropInfo(p)

        


    @precision.setter
    def precision(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：— \n
        数值精度
        '''
        self.set_prop('precision', value)

    @property
    def size(self):
        '''
        可选值：large/small \n
        计数器尺寸
        '''

        
        p = self.get_prop('size')
        return StrTypePropInfo(p)

        


    @size.setter
    def size(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：large/small \n
        计数器尺寸
        '''
        self.set_prop('size', value)

    @property
    def disabled(self):
        '''
        可选值：large/small \n
        是否禁用计数器
        '''

        
        p = self.get_prop('disabled')
        return BoolTypePropInfo(p)

        


    @disabled.setter
    def disabled(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large/small \n
        是否禁用计数器
        '''
        self.set_prop('disabled', value)

    @property
    def controls(self):
        '''
        可选值：large/small \n
        是否使用控制按钮
        '''

        
        p = self.get_prop('controls')
        return BoolTypePropInfo(p)

        


    @controls.setter
    def controls(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large/small \n
        是否使用控制按钮
        '''
        self.set_prop('controls', value)

    @property
    def controls_position(self):
        '''
        可选值：right \n
        控制按钮位置
        '''

        
        p = self.get_prop('controls-position')
        return StrTypePropInfo(p)

        


    @controls_position.setter
    def controls_position(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：right \n
        控制按钮位置
        '''
        self.set_prop('controls-position', value)

    @property
    def name(self):
        '''
        可选值：right \n
        原生 name 属性
        '''

        
        p = self.get_prop('name')
        return StrTypePropInfo(p)

        


    @name.setter
    def name(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：right \n
        原生 name 属性
        '''
        self.set_prop('name', value)

    @property
    def label(self):
        '''
        可选值：right \n
        输入框关联的 label 文字
        '''

        
        p = self.get_prop('label')
        return StrTypePropInfo(p)

        


    @label.setter
    def label(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：right \n
        输入框关联的 label 文字
        '''
        self.set_prop('label', value)

    @property
    def placeholder(self):
        '''
        可选值：right \n
        输入框默认 placeholder
        '''

        
        p = self.get_prop('placeholder')
        return StrTypePropInfo(p)

        


    @placeholder.setter
    def placeholder(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：right \n
        输入框默认 placeholder
        '''
        self.set_prop('placeholder', value)



from typing import Dict, List, TYPE_CHECKING
import pandas as pd
from pydantic import BaseModel

from pyvisflow.utils.helper import df2array_dict
from pyvisflow.models.TStaticData import TDataframe, TInfo, TColumnsType
from pyvisflow.core.components.file import FileData
# if TYPE_CHECKING:
#     from pyvisflow.core.components.file import FileData
    

def _2infos(df: pd.DataFrame) -> TInfo:
    cols = df.columns.tolist()
    rows = len(df)
    return TInfo(columns=cols, rows=rows)


def _2columnsType(df: pd.DataFrame) -> List[TColumnsType]:
    return [
        TColumnsType(name=name, type=str(t))
        for name, t in zip(df.columns, df.dtypes)
    ]


class TDataFrameInfo(BaseModel):
    id: str
    type: str 


class DataFrameManager():
    def __init__(self) -> None:
        self.df_dict: Dict[str, pd.DataFrame] = {}
        self.fileData_dict: Dict[str, FileData] = {}

    def mark_dataframe(self, data: pd.DataFrame):
        data_id = str(id(data))
        if not data_id in self.df_dict:
            self.df_dict[data_id] = data

        return TDataFrameInfo(id=data_id,type='dataframe')

    def mark_fileData(self, data:FileData):
        data_id = str(id(data))
        if not data_id in self.fileData_dict:
            self.fileData_dict[data_id] = data

        return TDataFrameInfo(id=data._id,type='from-file')

    def reset(self):
        self.df_dict.clear()

    def to_model(self) -> List[TDataframe]:

        return [
            TDataframe(id=id,
                       infos=_2infos(df),
                       columnsType=_2columnsType(df),
                       data=df2array_dict(df))
            for id, df in self.df_dict.items()
        ]



from typing import Union,Dict,List
from pyvisflow.core.components.components import Component
from pyvisflow.core.props import TypePropInfo,StrTypePropInfo,SubscriptableTypePropInfo, NumberTypePropInfo, BoolTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.core.props import *


class _Input(Component):
    '''
    Input 输入框
通过鼠标或键盘输入字符
    '''

    def __init__(self) -> None:
        super().__init__('input', TComponentType.builtIn)

    def _ex_get_react_data(self):
        return super()._ex_get_react_data()


    @property
    def type(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        类型
        '''

        
        p = self.get_prop('type')
        return StrTypePropInfo(p)

        


    @type.setter
    def type(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        类型
        '''
        self.set_prop('type', value)

    @property
    def value(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        绑定值
        '''

        
        p = self.get_prop('value')
        return StrTypePropInfo(p)

        


    @value.setter
    def value(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        绑定值
        '''
        self.set_prop('value', value)

    @property
    def maxlength(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        最大输入长度
        '''

        
        p = self.get_prop('maxlength')
        return NumberTypePropInfo(p)

        


    @maxlength.setter
    def maxlength(self, value: Union[NumberTypePropInfo,int]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        最大输入长度
        '''
        self.set_prop('maxlength', value)

    @property
    def minlength(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        原生属性，最小输入长度
        '''

        
        p = self.get_prop('minlength')
        return NumberTypePropInfo(p)

        


    @minlength.setter
    def minlength(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        原生属性，最小输入长度
        '''
        self.set_prop('minlength', value)

    @property
    def show_word_limit(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否显示输入字数统计，只在 type = "text" 或 type = "textarea" 时有效
        '''

        
        p = self.get_prop('show-word-limit')
        return BoolTypePropInfo(p)

        


    @show_word_limit.setter
    def show_word_limit(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否显示输入字数统计，只在 type = "text" 或 type = "textarea" 时有效
        '''
        self.set_prop('show-word-limit', value)

    @property
    def placeholder(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        输入框占位文本
        '''

        
        p = self.get_prop('placeholder')
        return StrTypePropInfo(p)

        


    @placeholder.setter
    def placeholder(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        输入框占位文本
        '''
        self.set_prop('placeholder', value)

    @property
    def clearable(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否可清空
        '''

        
        p = self.get_prop('clearable')
        return BoolTypePropInfo(p)

        


    @clearable.setter
    def clearable(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否可清空
        '''
        self.set_prop('clearable', value)

    @property
    def show_password(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否显示切换密码图标
        '''

        
        p = self.get_prop('show-password')
        return BoolTypePropInfo(p)

        


    @show_password.setter
    def show_password(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否显示切换密码图标
        '''
        self.set_prop('show-password', value)

    @property
    def disabled(self):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否禁用
        '''

        
        p = self.get_prop('disabled')
        return BoolTypePropInfo(p)

        


    @disabled.setter
    def disabled(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：text，textarea 和其他原生 input 的 type 值 \n
        是否禁用
        '''
        self.set_prop('disabled', value)

    @property
    def size(self):
        '''
        可选值：large / default / small \n
        输入框尺寸，只在 type !="textarea" 时有效
        '''

        
        p = self.get_prop('size')
        return StrTypePropInfo(p)

        


    @size.setter
    def size(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：large / default / small \n
        输入框尺寸，只在 type !="textarea" 时有效
        '''
        self.set_prop('size', value)

    @property
    def rows(self):
        '''
        可选值：large / default / small \n
        输入框行数，只对 type="textarea" 有效
        '''

        
        p = self.get_prop('rows')
        return NumberTypePropInfo(p)

        


    @rows.setter
    def rows(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：large / default / small \n
        输入框行数，只对 type="textarea" 有效
        '''
        self.set_prop('rows', value)

    @property
    def autosize(self):
        '''
        可选值：large / default / small \n
        textarea高度是否自适应，只在 type="textarea" 时生效。 可以接受一个对象，比如: { minRows: 2, maxRows: 6 }
        '''

        
        p = self.get_prop('autosize')
        return BoolTypePropInfo(p)

        


    @autosize.setter
    def autosize(self, value: Union[BoolTypePropInfo,Union[bool,Dict]]):
        '''
        可选值：large / default / small \n
        textarea高度是否自适应，只在 type="textarea" 时生效。 可以接受一个对象，比如: { minRows: 2, maxRows: 6 }
        '''
        self.set_prop('autosize', value)

    @property
    def autocomplete(self):
        '''
        可选值：large / default / small \n
        原生属性，自动补全
        '''

        
        p = self.get_prop('autocomplete')
        return StrTypePropInfo(p)

        


    @autocomplete.setter
    def autocomplete(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：large / default / small \n
        原生属性，自动补全
        '''
        self.set_prop('autocomplete', value)

    @property
    def name(self):
        '''
        可选值：large / default / small \n
        原生属性
        '''

        
        p = self.get_prop('name')
        return StrTypePropInfo(p)

        


    @name.setter
    def name(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：large / default / small \n
        原生属性
        '''
        self.set_prop('name', value)

    @property
    def readonly(self):
        '''
        可选值：large / default / small \n
        原生属性，是否只读
        '''

        
        p = self.get_prop('readonly')
        return BoolTypePropInfo(p)

        


    @readonly.setter
    def readonly(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large / default / small \n
        原生属性，是否只读
        '''
        self.set_prop('readonly', value)

    @property
    def max(self):
        '''
        可选值：large / default / small \n
        原生属性，设置最大值
        '''

        
        p = self.get_prop('max')
        return NumberTypePropInfo(p)

        


    @max.setter
    def max(self, value: Union[NumberTypePropInfo,int]):
        '''
        可选值：large / default / small \n
        原生属性，设置最大值
        '''
        self.set_prop('max', value)

    @property
    def min(self):
        '''
        可选值：large / default / small \n
        原生属性，设置最小值
        '''

        
        p = self.get_prop('min')
        return NumberTypePropInfo(p)

        


    @min.setter
    def min(self, value: Union[NumberTypePropInfo,int]):
        '''
        可选值：large / default / small \n
        原生属性，设置最小值
        '''
        self.set_prop('min', value)

    @property
    def step(self):
        '''
        可选值：large / default / small \n
        原生属性，设置输入字段的合法数字间隔
        '''

        
        p = self.get_prop('step')
        return NumberTypePropInfo(p)

        


    @step.setter
    def step(self, value: Union[NumberTypePropInfo,int]):
        '''
        可选值：large / default / small \n
        原生属性，设置输入字段的合法数字间隔
        '''
        self.set_prop('step', value)

    @property
    def resize(self):
        '''
        可选值：none / both / horizontal / vertical \n
        控制是否能被用户缩放
        '''

        
        p = self.get_prop('resize')
        return StrTypePropInfo(p)

        


    @resize.setter
    def resize(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：none / both / horizontal / vertical \n
        控制是否能被用户缩放
        '''
        self.set_prop('resize', value)

    @property
    def autofocus(self):
        '''
        可选值：none / both / horizontal / vertical \n
        原生属性，自动获取焦点
        '''

        
        p = self.get_prop('autofocus')
        return BoolTypePropInfo(p)

        


    @autofocus.setter
    def autofocus(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：none / both / horizontal / vertical \n
        原生属性，自动获取焦点
        '''
        self.set_prop('autofocus', value)

    @property
    def input_style(self):
        '''
        可选值：none / both / horizontal / vertical \n
        input 元素或 textarea 元素的 style
        '''

        
        return self.get_prop('input-style')
        


    @input_style.setter
    def input_style(self, value: Union[TypePropInfo,Dict]):
        '''
        可选值：none / both / horizontal / vertical \n
        input 元素或 textarea 元素的 style
        '''
        self.set_prop('input-style', value)



from typing import Union,Dict,List
from pyvisflow.core.components.components import Component
from pyvisflow.core.props import TypePropInfo,StrTypePropInfo,SubscriptableTypePropInfo, NumberTypePropInfo, BoolTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.core.props import *


class _EChart(Component):
    '''
    echart 图表
    '''

    def __init__(self) -> None:
        super().__init__('echart', TComponentType.builtIn)

    def _ex_get_react_data(self):
        return super()._ex_get_react_data()


    @property
    def title(self):
        '''
        可选值：nan \n
        图表标题
        '''

        
        p = self.get_prop('option.title.text')
        return StrTypePropInfo(p)

        


    @title.setter
    def title(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：nan \n
        图表标题
        '''
        self.set_prop('option.title.text', value)



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Generic, List, Optional, TypeVar, Union, Tuple, Dict, TYPE_CHECKING, Generator
from pyvisflow.core.props.indexProp import IndexPropInfo

from .absProp import AbsPropInfo

if TYPE_CHECKING:
    from pyvisflow.core.props.propValuator import Valuator


class NamePropInfo(AbsPropInfo):
    def __init__(self, valuator: Valuator, parent: Union[AbsPropInfo, None],
                 name: str) -> None:
        super().__init__(valuator, parent)
        self.name = name

    def _ex_gen_expr(self):
        return f'.{self.name}'



from .absProp import *
from .indexProp import *
from .methodProp import *
from .nameProp import *
from .typeProp import *
from .valueProp import *
from .propValuator import *
from .dataFilterProp import *


from __future__ import annotations
from typing import Dict, Optional, TypeVar, Union, TYPE_CHECKING
from pyvisflow.core.props import BoolTypePropInfo
from pyvisflow.core.props.fnsMethodProp import FnsMethodPropInfo
from pyvisflow.core.props.typeProp import NumberTypePropInfo, StrTypePropInfo, TypePropInfo
from pyvisflow.core.props.valueProp import DictValuePropInfo, ValuePropInfo

TIfelse = TypeVar('TIfelse', str, int, Dict)
TMapReturn = TypeVar('TMapReturn', str, int, TypePropInfo)

if TYPE_CHECKING:
    from pyvisflow.app import App


class Fns():
    def __init__(self, app: Optional[App] = None) -> None:
        self.__app = app

    def ifelse(self, condition: BoolTypePropInfo,
               value_if_true: Optional[TIfelse],
               value_if_false: Optional[TIfelse]):

        value_true = ValuePropInfo.try2PropInfo(value_if_true)
        value_false = ValuePropInfo.try2PropInfo(value_if_false)

        m = FnsMethodPropInfo('ifelse', [condition, value_true, value_false])
        if isinstance(value_if_true, str):
            return StrTypePropInfo(m)

        return NumberTypePropInfo(m)

    def map(self,
            key: StrTypePropInfo,
            data: Dict[str, TMapReturn],
            __default=None) -> StrTypePropInfo:
        assert len(data) > 0, 'dict data len must be greater than 0'

        data_value = DictValuePropInfo(None, data)
        m = FnsMethodPropInfo('map', [key, data_value, __default])

        first = list(data.values())[0]
        if isinstance(first, (str, StrTypePropInfo)):
            return StrTypePropInfo(m)

        return NumberTypePropInfo(m)


from __future__ import annotations

from abc import abstractmethod
from typing import Callable, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, Union

if TYPE_CHECKING:
    from pyvisflow.core.props.propValuator import Valuator


class AbsPropInfo():
    def __init__(self, valuator: Valuator,
                 parent: Union[AbsPropInfo, None]) -> None:
        super().__init__()
        self.valuator = valuator
        self.parent = parent

    @abstractmethod
    def _ex_gen_expr(self) -> str:
        return ''

    def gen_expr(self):
        if self.parent:
            return f'{self.parent.gen_expr()}{self._ex_gen_expr()}'
        return self._ex_gen_expr()



from typing import Callable, Dict, Hashable, List, Optional, Set, Tuple, Union
from pydantic import BaseModel, PrivateAttr, Field
from enum import Enum


class TTargetWatchInfo(BaseModel):
    id: str
    path: str
    logic: str


class TLogic(BaseModel):
    logic: str


class TChartFilterTarget(BaseModel):
    id: str
    logic: str


class TTableFilterTarget(BaseModel):
    id: str
    logic: str


class TValueWatch(BaseModel):
    target: TTargetWatchInfo


class TChartFilterWatch(BaseModel):
    target: TChartFilterTarget


class TTableFilterWatch(BaseModel):
    target: TTableFilterTarget


class TWatchInfo(BaseModel):
    values: List[TValueWatch] = []
    chartFilters: List[TChartFilterWatch] = []
    tableFilters: List[TTableFilterWatch] = []



from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, List, TYPE_CHECKING, Optional, Union

import pandas as pd
from pyvisflow.core.components.components import Component
from pyvisflow.core.components.dataTable import DataTable
from pyvisflow.core.components.echart import EChart
from pyvisflow.core.components.icon import Icon
from pyvisflow.core.components.input import Input
from pyvisflow.core.components.inputNumber import InputNumber
from pyvisflow.core.components.plotly import Plotly
from pyvisflow.core.components.select import Select
from pyvisflow.core.components.text import Text
from pyvisflow.core.components.file import File

from pyvisflow.core.layout import MarkdownLayout
from pyvisflow.core.props import BoolTypePropInfo

from pyvisflow.models.TComponent import TComponentType

if TYPE_CHECKING:
    from pyvisflow.core.props import TypePropInfo
    from pyvisflow.core.components.file import FileData


class Container(Component):
    # appender_stack: List[Callable[[Component]
    def __init__(self, tag: str, type=TComponentType.builtIn) -> None:
        super().__init__(tag, type)
        # self._appender_stack = appender_stack

    def add_child(self, cp: Component):
        self._children.append(cp)
        return cp


class BoxContainer(Container):
    def __init__(self, sepc=0) -> None:
        super().__init__('box')
        self._spec = sepc
        self._on_before_data_fn = lambda data: data
        self._on_before_add_child_fn = BoxContainer.__default_on_before_add_child_fn

    @staticmethod
    def __default_on_before_add_child_fn(cp: Component):
        pass

    def file(self):
        cp = File()
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def box(self, ):
        cp = BoxContainer()
        cp._on_before_data_fn = self._on_before_data_fn
        cp._on_before_add_child_fn = self._on_before_add_child_fn

        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def row(self, ):
        cp = Row()
        cp._on_before_data_fn = self._on_before_data_fn
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def icon(self, svg: str, size: str):
        cp = Icon(svg, size)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def icon_fromfile(self, file: str, size: str):
        with open(file, 'r', encoding='utf8') as f:
            return self.icon(f.read(), size)

    def text(self, value: str):
        cp = Text(value)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def dataTable(self, dataframe: Union[pd.DataFrame,FileData], page_size=5):
        cp = DataTable(self._on_before_data_fn(dataframe))
        cp.page_size = page_size
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def tabs(self, names: Union[Iterable[str], Dict[str, str]]):
        cp = tabs.Tabs(names)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)

        for box in cp.boxes:
            box._on_before_data_fn = self._on_before_data_fn
            box._on_before_add_child_fn = self._on_before_add_child_fn

        return cp

    def plotly(self):
        cp = Plotly()
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def cols(self, spec: Union[int, List[int]]):
        cp = ColumnsContainer(spec)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)

        for box in cp.boxes:
            box._on_before_data_fn = self._on_before_data_fn
            box._on_before_add_child_fn = self._on_before_add_child_fn

        return cp.boxes

    def input(self, value=''):
        cp = Input(value)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def inputNumber(self, value=0):
        cp = InputNumber(value)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def select(self, options: Iterable[str]):
        cp = Select(options)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def echart(self, data:Optional[pd.DataFrame]=None):
        info=None
        if not data is None:
            info = self._on_before_data_fn(data)
        cp = EChart(info)
        self._on_before_add_child_fn(cp)
        self.add_child(cp)
        return cp

    def markdown(self, md: str, **prop_mapping: TypePropInfo):
        layout = MarkdownLayout(md, prop_mapping)

        for cp in layout.parse_markdown():
            self._on_before_add_child_fn(cp)
            self.add_child(cp)
        return layout

    @property
    def isHover(self):
        p = self.get_prop('isHover')
        return BoolTypePropInfo(p)

    def _ex_get_react_data(self) -> Dict[str, Any]:
        data = super()._ex_get_react_data()
        data.update({'spec': self._spec, 'isHover': False})
        return data


class ColumnsContainer(Container):
    def __init__(
            self,
            #  appender_stack: List[Callable[[Component], Component]],
            spec: Union[int, List[int]]) -> None:
        super().__init__('columns-container')

        self._is_grid = not isinstance(spec, int)

        __spec = spec
        if isinstance(spec, int):
            __spec = [spec for _ in range(spec)]
        self._spec = __spec
        self._boxes = tuple([BoxContainer(s) for s in __spec])
        for box in self._boxes:
            self.add_child(box)

    @property
    def boxes(self):
        return self._boxes

    def _ex_get_react_data(self) -> Dict[str, Any]:
        data = super()._ex_get_react_data()
        data.update({'spec': self._spec, 'isGrid': self._is_grid})
        return data


class Row(BoxContainer):
    def __init__(self) -> None:
        super().__init__()
        self._tag='row'

    @property
    def isWrap(self):
        '''

        '''

        p = self.get_prop('isWrap')
        return BoolTypePropInfo(p)

    @isWrap.setter
    def isWrap(self, value: Union[BoolTypePropInfo, bool]):
        '''
        '''
        self.set_prop('isWrap', value)


import pyvisflow.core.components.tabs as tabs


from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union
from pyvisflow.core.dataManager import TDataFrameInfo
from pyvisflow.core.props import SubscriptableTypePropInfo, StrTypePropInfo, NumberTypePropInfo
from pyvisflow.core.props.dataFilterProp import DataFilterPropInfo
from pyvisflow.core.props.nameProp import NamePropInfo
from pyvisflow.core.props.typeProp import BoolTypePropInfo, TypePropInfo
from pyvisflow.core.reactive import Reactive
from pyvisflow.models.TComponent import TComponentType
import pandas as pd

from pyvisflow.models.TWatchInfo import TChartFilterWatch, TChartFilterTarget
from pyvisflow.utils.helper import df2array_dict
from .auto_create._echart import _EChart


class EChart(_EChart):
    def __init__(
            self,
            data_info: Optional[TDataFrameInfo]=None,
    ) -> None:
        super().__init__()
        self._data_info = data_info
        self._series = []
        self.update_styles(height='400px', width='100%')
        self._filter_watch_infos: List[TChartFilterWatch] = []
        (
            self.set_option('dataset',{
                    'dimensions': [],
                        'source': {}
            })
            .set_option('xAxis',{}).set_option('yAxis',{})
        )

    @property
    def utils(self, ):
        return Helper(self)


    def set_option(self, prop: str, data: Union[Any, TypePropInfo]):
        if not prop.startswith('option.'):
            prop = f'option.{prop}'
        self.set_prop(prop, data)
        return self

    def set_series(self, idx: int, prop: str, data: Union[Any, TypePropInfo]):
        self.set_option(f'series[{idx}].{prop}', data)
        return self

    @property
    def clickInfo(self):
        p = self.get_prop('clickInfo')
        return SubscriptableTypePropInfo[str, StrTypePropInfo](p)

    @property
    def data(self):
        p = DataFilterPropInfo('row')
        return SubscriptableTypePropInfo[str, StrTypePropInfo](p)

    def filters(self, condition: BoolTypePropInfo):
        logic = condition.valuator.cal(condition)
        target = TChartFilterTarget(id=self._id, logic=logic)
        filter = TChartFilterWatch(target=target)
        self._filter_watch_infos.append(filter)

    def _ex_get_react_data(self):
        data = super()._ex_get_react_data()

        info_id = None
        if self._data_info:
            info_id = self._data_info.id

        data.update({
            'dataInfo': {
                'id': info_id
            },
            'option':{

            },
            'clickInfo': {},
        })
        return data


class Helper():
    def __init__(self, echart: EChart) -> None:
        self._echart = echart

    def use_from_option_dict(self,option:Dict):
        for key,value in option.items():
            self._echart.set_option(key,value)
        return self

    def use_pie(self, itemName: str, value: str):
        self._echart.set_option('series', [{
            'type': 'pie',
            'encode': {
                'itemName': itemName,
                'value': value
            },
            'id': self._echart._id,
            "universalTransition": True,
        }]) \
        .set_option('legend', {
                                'top': '5%',
                                'left': 'center'
                            },)
        return PieHelper(self._echart)

    def use_bar(self, x: str, y: str):
        self._echart.set_option('series', [{
            'type': 'bar',
            'showBackground': True,
            'backgroundStyle': {
                'color': 'rgba(180, 180, 180, 0.2)'
            },
            'encode': {
                'x': x,
                'y': y
            },
            'id': self._echart._id,
            "universalTransition": True,
        }]).set_option('xAxis', {'type': 'category'})
        return BarHelper(self._echart)

    def use_line(self, x: str, y: str, color: Optional[str] = None):
        self._echart.set_option('series', [{
            'type': 'line',
            'encode': {
                'x': x,
                'y': y,
                'color' :color
            },
            'id': self._echart._id,
            "universalTransition": True,
        }]) \
        .set_option('xAxis', {'type': 'category'}) \
        .set_option('yAxis', {'type': 'value'}) \
        .set_option('tooltip', {'trigger': 'axis'})
        return LineHelper(self._echart)


class LineHelper():
    def __init__(self, echart: EChart) -> None:
        self._echart = echart

    def area_chart(self):
        '''
        基础面积图
        '''
        self._echart \
                    .set_series(0, 'areaStyle', {}) \
                    .set_series(0,'symbol','none') \
                    .set_series(0,'sampling','lttb') \
                    .set_series(0,  "lineStyle",{"width": 0}) \
                    .set_option('xAxis', {'type': 'category'})

        return self

    def smooth(self):
        '''
        平滑化
        '''
        self._echart \
                    .set_series(0, 'smooth', True) \

        return self


class BarHelper():
    def __init__(self, echart: EChart) -> None:
        self._echart = echart

    def showBackground(self, color: str = 'rgba(180, 180, 180, 0.2)'):
        '''
        显示柱子背景
        '''
        self._echart \
                    .set_series(0, 'showBackground', True) \
                    .set_series(0, 'backgroundStyle',  {
                        'color': color
                    }) \

        return self


class PieHelper():
    def __init__(self, echart: EChart) -> None:
        self._echart = echart

    def circle_chart(self):
        '''
        圆环图
        '''
        self._echart.set_series(0, 'labelLine', {
            'show': False
        }).set_series(0, 'emphasis', {
            'label': {
                'show': True,
                'fontSize': '40',
                'fontWeight': 'bold'
            }
        }).set_series(0, 'label', {
            'show': False,
            'position': 'center'
        }).set_series(0, 'itemStyle', {
            'borderRadius': 10,
            'borderColor': '#fff',
            'borderWidth': 2
        }).set_series(0, 'avoidLabelOverlap', False).set_series(
            0, 'radius', ['40%', '70%']).set_series(0, 'type', 'pie')

        return self

    def nightingale_rose_chart(self):
        '''
        南丁格尔玫瑰图
        '''
        self._echart \
                    .set_series(0, 'type', 'pie') \
                    .set_series(0, 'center',  ['50%', '50%']) \
                    .set_series(0, 'roseType', 'area') \
                    .set_series(0, 'itemStyle', {
                                                'borderRadius': 8
                                            }) \

        return self


from __future__ import annotations

from abc import abstractmethod
from typing import Callable, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator

from .absProp import AbsPropInfo
from pyvisflow.core.props.propValuator import CombineValuator
if TYPE_CHECKING:
    from .propValuator import Valuator


class CombinePropInfo(AbsPropInfo):
    def __init__(self, logic: str, left: AbsPropInfo,
                 right: AbsPropInfo) -> None:
        super().__init__(CombineValuator(), None)
        self.left = left
        self.right = right
        self.logic = logic

    def _ex_gen_expr(self):
        left = self.left.valuator.cal(self.left)
        right = self.right.valuator.cal(self.right)
        return f'({left}{self.logic}{right})'



from typing import Any, Dict, List, Optional
from pydantic import BaseModel
from enum import Enum


class TInfo(BaseModel):
    columns: List[str]
    rows: int


class TColumnsType(BaseModel):
    name: str
    type: str


class TDataframe(BaseModel):
    id: str
    infos: TInfo
    columnsType: List[TColumnsType]
    data: List[List[Any]]


class TStaticData(BaseModel):
    dataframes: List[TDataframe]



from typing import Dict, List, Iterable, Union

import pyvisflow.core.components.containers as containers
from .auto_create._tabs import _Tabs


class Tabs(_Tabs):
    def __init__(self, names: Union[Iterable[str], Dict[str, str]]) -> None:
        super().__init__()
        labels = []
        if isinstance(names, Iterable):
            names = [str(n) for n in names]
            labels = names

        if isinstance(names, Dict):
            labels = list(str(n) for n in names.values())
            names = list(str(n) for n in names.keys())

        self._labels = labels
        self.names = names

        self._boxes = tuple(
            [containers.BoxContainer() for _ in range(len(names))])

        self._mapping = {name: box for name, box in zip(names, self._boxes)}
        for box in self._boxes:
            self.add_child(box)

    @property
    def boxes(self):
        return self._boxes

    def __iter__(self):
        return iter(self.boxes)

    def __getitem__(self, key: Union[str, int]):
        if isinstance(key, int):
            return self._boxes[key]
        return self._mapping[key]

    def _ex_get_react_data(self):
        data = super()._ex_get_react_data()
        data.update({
            'labels': self._labels,
        })
        return data



from datetime import datetime
from pathlib import Path
from json import JSONEncoder, dumps
from functools import partial
import numpy as np
import pandas as pd

_global_id = 0


def get_global_id():
    global _global_id
    _global_id += 1
    return str(_global_id)


_m_project_root = Path(__file__).absolute().parent.parent


def get_project_root():
    return _m_project_root


def _nan2None(obj):



    if isinstance(obj, dict):
        return {k: _nan2None(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_nan2None(v) for v in obj]
    elif isinstance(obj, float) and pd.isna(obj):
        return None
    elif isinstance(obj,np.ndarray):
        return obj.tolist()
    elif isinstance(obj,datetime):
        return str(obj)

    return obj


class _NanConverter(JSONEncoder):
    def default(self, obj):
        # possible other customizations here
        pass

    def encode(self, obj, *args, **kwargs):
        obj = _nan2None(obj)
        return super().encode(obj, *args, **kwargs)

    def iterencode(self, obj, *args, **kwargs):
        obj = _nan2None(obj)
        return super().iterencode(obj, *args, **kwargs)


json_dumps_fn = partial(dumps, cls=_NanConverter)


from typing import Any, Dict
from pyvisflow.models.TComponent import TComponentType
from .components import Component
from bs4 import BeautifulSoup


class Icon(Component):
    def __init__(self, svg: str, size: str) -> None:
        super().__init__('icon', TComponentType.builtIn)

        s = BeautifulSoup(svg, 'html.parser')

        del s.find('svg')['width']
        del s.find('svg')['height']
        self._svg = str(s)
        self._size = size

    def _ex_get_react_data(self) -> Dict[str, Any]:
        data = super()._ex_get_react_data()
        data.update({'svg': self._svg, 'size': self._size})
        return data


from typing import Dict, Hashable, Any, Optional, Union
from pyvisflow.core.reactive import Reactive
from pyvisflow.core.props import TypePropInfo


class Structure(Reactive):
    def __init__(self) -> None:
        super().__init__()


class Mapping(Structure):
    def __init__(self,
                 mapping: Optional[Dict[str, Dict[str, Any]]] = None) -> None:
        super().__init__()
        self._mapping = mapping or {}
        self.key = ''

    def add_item(self, key: str, datas: Union[Dict[str, Any], Any]):
        if not isinstance(datas, dict):
            datas = {'__value__': datas}

        self._mapping.update({key: datas})

        return self

    @property
    def value(self):
        return self['__value__']

    @property
    def key(self):
        return self.get_prop('key')

    @key.setter
    def key(self, value: Union[TypePropInfo, str]):
        self.set_prop('key', value)

    # def __getitem__(self, item: str):
    #     return (ExPropInfo(
    #         self, f'key').set_watchingPath(f"`mapping['${{change}}'].{item}`"))

    def _ex_get_react_data(self) -> Dict[str, Any]:
        data = super()._ex_get_react_data()
        data.update({'mapping': self._mapping})
        return data



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, Union

from pyvisflow.utils.helper import value2code

from .absProp import AbsPropInfo
from pyvisflow.core.props.propValuator import FnsValuator

if TYPE_CHECKING:
    from pyvisflow.core.props.typeProp import TypePropInfo


class FnsMethodPropInfo(AbsPropInfo):
    def __init__(self,
                 name: str,
                 args: Optional[List] = None,
                 package_var='fns') -> None:
        super().__init__(FnsValuator(package_var), None)
        self.name = name
        self.args = args or []

    @staticmethod
    def __to_code(v: Union[str, int, TypePropInfo]):
        if (v is None) or isinstance(v, (str, int,List)):
            return value2code(v)
        return v.valuator.cal(v)

    def _ex_gen_expr(self):
        args = ','.join(FnsMethodPropInfo.__to_code(v) for v in self.args)
        return f'.{self.name}({args})'



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union,TYPE_CHECKING
from pyvisflow.core.components.containers import BoxContainer, Component
from pyvisflow.core.components import DataTable, EChart
from pyvisflow.core.dataManager import DataFrameManager
from pyvisflow.core.props.fns import Fns
from pyvisflow.core.structure import Mapping, Structure
from pyvisflow.models.TConfig import TReactData, TConfig
from pyvisflow.models.TStaticData import TStaticData
from pyvisflow.models.TWatchInfo import TWatchInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.utils.data_gen import get_project_root, json_dumps_fn
from pyvisflow.utils.helper import flatten
import pandas as pd
import re
from pyvisflow.core.components.file import FileData
if TYPE_CHECKING:
    from pyvisflow.core.reactive import Reactive
    

__scripts_mapping__ = {'plotly': 'plotly-2.9.0.min.js'}
__re_app_div__ = re.compile(r'<div\sid="app"></div>')


class App(BoxContainer):
    def __init__(self) -> None:
        super().__init__()
        self.structures: List[Reactive] = []
        self.df_manager = DataFrameManager()
        self._scripts: Set[str] = set()
        self.fns = Fns(self)

        def on_before_data_fn(data: Union[pd.DataFrame,FileData]):
            if isinstance(data, pd.DataFrame):
                return self.df_manager.mark_dataframe(data)

            if isinstance(data,FileData):
                self.structures.append(data)
                return self.df_manager.mark_fileData(data)

            return data

        self._on_before_data_fn = on_before_data_fn

        def on_before_add_child_fn(cp: Component):
            if cp._tag == 'plotly':
                self._scripts.add(cp._tag)

        self._on_before_add_child_fn = on_before_add_child_fn

    @property
    def Fns(self):
        return self.fns

    def mapping(self, mapping: Optional[Dict[str, Dict[str, Any]]] = None):
        m = Mapping(mapping)
        self.structures.append(m)
        return m

    def add_srcipt(self, *names: str):
        for n in names:
            assert n in __scripts_mapping__, f'not support {n} script'
        for n in names:
            self._scripts.add(n)

    def pretty_json(self):
        return json_dumps_fn(self.create_config().dict(), indent=2)

    def config2file(self, file):
        return Path(file).write_text(self.pretty_json(), 'utf8')

    def create_config(self):
        cps = [self.to_model()]

        # reactData for each
        cps_flatten = list(flatten(self, lambda x: x._children))
        reactData = [
            TReactData(id=cp._id, data=cp._create_react_data())
            for cp in cps_flatten if cp._type not in
            {TComponentType.htmlRaw, TComponentType.htmlRawString}
        ]

        structure_data = [
            TReactData(id=s._id, data=s._create_react_data())
            for s in self.structures
        ]

        reactData.extend(structure_data)

        valueInfos = []
        tableInfos = []
        chartInfos = []
        for cp in cps_flatten:
            valueInfos.extend(cp._get_WatchInfos())
            if isinstance(cp, DataTable):
                tableInfos.extend(cp._filter_watch_infos)
            if isinstance(cp, EChart):
                chartInfos.extend(cp._filter_watch_infos)

        for s in self.structures:
            valueInfos.extend(s._get_WatchInfos())

        watchInfo = TWatchInfo(values=valueInfos,
                               tableFilters=tableInfos,
                               chartFilters=chartInfos)

        return TConfig(
            staticData=TStaticData(dataframes=self.df_manager.to_model()),
            cps=cps,
            reactDatas=reactData,
            watchInfo=watchInfo)

    def __reset_data(self):
        self._children.clear()
        self.df_manager.reset()

    def to_html(self, file):
        file = Path(file)
        symbol = '"__{{__config_data__}}___"'
        scripts_tag = '<!-- [__other_scripts__] -->'

        def script2module(name: str):
            file = get_project_root() / 'template' / __scripts_mapping__[name]
            return f'<script>{file.read_text()}</script>'

        script_codes = ' '.join(script2module(s) for s in self._scripts)

        config = self.create_config()
        self.__reset_data()
        config = json_dumps_fn(config.dict())

        with open(get_project_root() / 'template/index.html',
                  mode='r',
                  encoding='utf8') as html:
            res = html.read().replace(symbol, config).replace(
                '<div id="app"></div>', f'<div id="app"></div> {script_codes}')
            # res = __re_app_div__.sub(f'<div id="app"></div> {script_codes}',
            #                          res)

            Path(file).write_text(res, 'utf8')

        print(f'to html:{file.absolute()}')



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, Union
from pyvisflow.core.props.propValuator import FnsValuator

from pyvisflow.utils.helper import value2code

from .absProp import AbsPropInfo

if TYPE_CHECKING:
    from pyvisflow.core.props.propValuator import Valuator


class MethodPropInfo(AbsPropInfo):
    def __init__(self,
                 valuator: Valuator,
                 parent: Union[AbsPropInfo, None],
                 name: str,
                 args: Optional[List] = None) -> None:
        super().__init__(valuator, parent)
        self.name = name
        self.args = args or []

    def _ex_gen_expr(self):
        args = ','.join(value2code(v) for v in self.args)
        return f'?.{self.name}({args})'


class LambdaMethodPropInfo(AbsPropInfo):
    def __init__(self, value: AbsPropInfo,args:Optional[List[str]]=None) -> None:

        var = f"({','.join(args or [])})=>"

        super().__init__(FnsValuator(var), None)
        self.value = value

    def _ex_gen_expr(self):
        return self.value.valuator.cal(self.value)


from typing import Dict, List, Optional
from pydantic import BaseModel

from pyvisflow.models.TWatchInfo import TWatchInfo
from .TComponent import TComponent
from .TStaticData import TStaticData

from pyvisflow.utils.data_gen import json_dumps_fn

class TReactData(BaseModel):
    id: str
    data: Dict


class TConfig(BaseModel):
    staticData: TStaticData
    cps: List[TComponent]
    reactDatas: Optional[List[TReactData]] = []
    watchInfo: TWatchInfo


    class Config:
        json_dumps = json_dumps_fn




from typing import Union
from pyvisflow.core.components.components import Component
from pyvisflow.models.TComponent import TComponentType
from .auto_create._inputNumber import _InputNumber


class InputNumber(_InputNumber):
    def __init__(self, value: float) -> None:
        super().__init__()

        self.value = value



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Generic, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, TypeVar, Union
from pyvisflow.core.props.nameProp import NamePropInfo
from pyvisflow.core.props.valueProp import ValuePropInfo

from pyvisflow.utils.helper import value2code

from .absProp import AbsPropInfo
from .combineProp import CombinePropInfo
from .methodProp import MethodPropInfo
from .indexProp import IndexPropInfo
from .fnsMethodProp import FnsMethodPropInfo

if TYPE_CHECKING:
    from pyvisflow.core.props.propValuator import Valuator


class TypePropInfo(AbsPropInfo):
    def __init__(self, parent: AbsPropInfo) -> None:
        super().__init__(parent.valuator, parent)


class StrTypePropInfo(TypePropInfo):
    def __init__(self, parent: AbsPropInfo) -> None:
        super().__init__(parent)

    def __eq__(self, __o: Union[str, StrTypePropInfo]):
        v = None
        if isinstance(__o, str):
            v = ValuePropInfo(None, __o)
        else:
            v = __o
        cb = CombinePropInfo('==', self, v)
        return BoolTypePropInfo(cb)

    def __ne__(self, __o: StrTypePropInfo):
        cb = CombinePropInfo('!=', self, __o)
        return BoolTypePropInfo(cb)

    @staticmethod
    def __try2PropInfo(value:Union[str, StrTypePropInfo]):
        if isinstance(value,StrTypePropInfo):
            return value
        val = None
        if isinstance(value, str):
            val = ValuePropInfo(None, value)
        else:
            assert isinstance(
                value, TypePropInfo
            ), f'other type is [{type(value)}],it must be str, StrTypePropInfo'
            val = ValuePropInfo.try2PropInfo(value)

        return val

    @staticmethod
    def __Combine2str(logic: str,left:Union[str, StrTypePropInfo],right:Union[str, StrTypePropInfo]):
        left = StrTypePropInfo.__try2PropInfo(left)
        right = StrTypePropInfo.__try2PropInfo(right)
        cb = CombinePropInfo('+', left, right)
        return StrTypePropInfo(cb)

    def __add__(self, other: Union[str, StrTypePropInfo]):
        return StrTypePropInfo.__Combine2str('+',self,other)

    def __radd__(self, other: Union[str, StrTypePropInfo]):
        return StrTypePropInfo.__Combine2str('+',other,self)

    # def __int__(self) -> int:
    #     return self.toNumber()

    # def __float__(self) -> :
    #     return self.toNumber()

    def slice(self, start: int, end: Optional[int] = None):
        m = MethodPropInfo(self.valuator, self, 'slice', [start, end])

        res = StrTypePropInfo(m)
        return res

    def toNumber(self):
        m = FnsMethodPropInfo('toNumber', [self])
        return NumberTypePropInfo(m)

    def length(self):
        m = NamePropInfo(self.valuator, self, 'length')
        res = NumberTypePropInfo(m)
        return res

    def _ex_gen_expr(self):
        return ''


class NumberTypePropInfo(TypePropInfo):
    def __init__(self, parent: AbsPropInfo) -> None:
        super().__init__(parent)

    def __eq__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4comp('==', other)

    def __ne__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4comp('!=', other)

    def __lt__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4comp('<', other)

    def __le__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4comp('<=', other)

    def __gt__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4comp('>', other)

    def __ge__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4comp('>=', other)

    def __combine(self, logic: str,
                  other: Union[float, int, NumberTypePropInfo]):
        val = None
        if isinstance(other, (float, int)):
            val = ValuePropInfo(None, other)
        else:
            assert isinstance(
                other, TypePropInfo
            ), f'other type is [{type(other)}],it must be float,int'
            val = ValuePropInfo.try2PropInfo(other)

        return CombinePropInfo(logic, self, val)

    def __combine4comp(self, logic: str,
                       other: Union[float, int, NumberTypePropInfo]):
        cb = self.__combine(logic, other)
        return BoolTypePropInfo(cb)

    def __combine4cal(self, logic: str,
                      other: Union[float, int, NumberTypePropInfo]):
        cb = self.__combine(logic, other)
        return NumberTypePropInfo(cb)

    def __add__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4cal('+', other)

    def __radd__(self, other: Union[float, int, NumberTypePropInfo]):
        return self + other

    def __sub__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4cal('-', other)

    def __rsub__(self, other: Union[float, int, NumberTypePropInfo]):
        return self - other

    def __mul__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4cal('*', other)

    def __rmul__(self, other: Union[float, int, NumberTypePropInfo]):
        return self * other

    def __truediv__(self, other: Union[float, int, NumberTypePropInfo]):
        return self.__combine4cal('/', other)

    def __rtruediv__(self, other: Union[float, int, NumberTypePropInfo]):
        return self / other

    def __floordiv__(self, other: Union[float, int, NumberTypePropInfo]):
        oper1 = self.__combine4cal('/', other)
        m = FnsMethodPropInfo('floor', [oper1], package_var='Math')
        return m

    def __rfloordiv__(self, other: Union[float, int, NumberTypePropInfo]):
        return self // other

    def toString(self):
        m = MethodPropInfo(self.valuator,self,'toString')
        return StrTypePropInfo(m)

    def _ex_gen_expr(self):
        return ''


class BoolTypePropInfo(TypePropInfo):
    def __init__(self, parent: AbsPropInfo) -> None:
        super().__init__(parent)

    def __and__(self, other: BoolTypePropInfo):
        cb = CombinePropInfo('&&', self, other)
        return BoolTypePropInfo(cb)

    def __or__(self, other: BoolTypePropInfo):
        cb = CombinePropInfo('||', self, other)
        return BoolTypePropInfo(cb)

    def ex_gen_expr(self):
        return ''


T = TypeVar('T', str, int)
TItem = TypeVar('TItem', StrTypePropInfo, NumberTypePropInfo)


class SubscriptableTypePropInfo(TypePropInfo, Generic[T, TItem]):
    def __init__(self, parent: AbsPropInfo) -> None:
        super().__init__(parent)

    def __getitem__(self, item: Union[T, TItem]) -> TItem:
        idx = IndexPropInfo(self.valuator, self, item)

        if hasattr(self, '__orig_class__'):
            if self.__orig_class__.__args__[1] == StrTypePropInfo:
                return StrTypePropInfo(idx)

            if self.__orig_class__.__args__[1] == NumberTypePropInfo:
                return NumberTypePropInfo(idx)

        return SubscriptableTypePropInfo(idx)



from typing import Union,Dict,List
from pyvisflow.core.components.components import Component
from pyvisflow.core.props import TypePropInfo,StrTypePropInfo,SubscriptableTypePropInfo, NumberTypePropInfo, BoolTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.core.props import *


class _DataTable(Component):
    '''
    Input Number 数字输入框
仅允许输入标准的数字值，可定义范围
    '''

    def __init__(self) -> None:
        super().__init__('dataTable', TComponentType.builtIn)

    def _ex_get_react_data(self):
        return super()._ex_get_react_data()


    @property
    def current_page(self):
        '''
        可选值：nan \n
        分页时，当前页数
        '''

        
        p = self.get_prop('currentPage')
        return NumberTypePropInfo(p)

        


    @current_page.setter
    def current_page(self, value: Union[NumberTypePropInfo,int]):
        '''
        可选值：nan \n
        分页时，当前页数
        '''
        self.set_prop('currentPage', value)

    @property
    def show_columns(self):
        '''
        可选值：nan \n
        表格显示的列。当你希望页面表格只输出数据一部分列的时候可以设置此属性
        '''

        
        p = self.get_prop('showColumns')
        return SubscriptableTypePropInfo[int,StrTypePropInfo](p)

        


    @show_columns.setter
    def show_columns(self, value: Union[SubscriptableTypePropInfo[int,StrTypePropInfo],List[str]]):
        '''
        可选值：nan \n
        表格显示的列。当你希望页面表格只输出数据一部分列的时候可以设置此属性
        '''
        self.set_prop('showColumns', value)

    @property
    def page_size(self):
        '''
        可选值：nan \n
        分页时，每页的记录数
        '''

        
        p = self.get_prop('pageSize')
        return NumberTypePropInfo(p)

        


    @page_size.setter
    def page_size(self, value: Union[NumberTypePropInfo,int]):
        '''
        可选值：nan \n
        分页时，每页的记录数
        '''
        self.set_prop('pageSize', value)

    @property
    def height(self):
        '''
        可选值：nan \n
        Table 的高度， 默认为自动高度。 如果 height 为 number 类型，单位 px；如果 height 为 string 类型，则这个高度会设置为 Table 的 style.height 的值，Table 的高度会受控于外部样式。
        '''

        
        p = self.get_prop('height')
        return StrTypePropInfo(p)

        


    @height.setter
    def height(self, value: Union[StrTypePropInfo,Union[str,float]]):
        '''
        可选值：nan \n
        Table 的高度， 默认为自动高度。 如果 height 为 number 类型，单位 px；如果 height 为 string 类型，则这个高度会设置为 Table 的 style.height 的值，Table 的高度会受控于外部样式。
        '''
        self.set_prop('height', value)

    @property
    def max_height(self):
        '''
        可选值：nan \n
        Table 的最大高度。 合法的值为数字或者单位为 px 的高度。
        '''

        
        p = self.get_prop('max-height')
        return StrTypePropInfo(p)

        


    @max_height.setter
    def max_height(self, value: Union[StrTypePropInfo,Union[str,float]]):
        '''
        可选值：nan \n
        Table 的最大高度。 合法的值为数字或者单位为 px 的高度。
        '''
        self.set_prop('max-height', value)

    @property
    def stripe(self):
        '''
        可选值：nan \n
        是否为斑马纹 table
        '''

        
        p = self.get_prop('stripe')
        return BoolTypePropInfo(p)

        


    @stripe.setter
    def stripe(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：nan \n
        是否为斑马纹 table
        '''
        self.set_prop('stripe', value)

    @property
    def border(self):
        '''
        可选值：nan \n
        是否带有纵向边框
        '''

        
        p = self.get_prop('border')
        return BoolTypePropInfo(p)

        


    @border.setter
    def border(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：nan \n
        是否带有纵向边框
        '''
        self.set_prop('border', value)

    @property
    def size(self):
        '''
        可选值：large / default /small \n
        Table 的尺寸
        '''

        
        p = self.get_prop('size')
        return StrTypePropInfo(p)

        


    @size.setter
    def size(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：large / default /small \n
        Table 的尺寸
        '''
        self.set_prop('size', value)

    @property
    def fit(self):
        '''
        可选值：large / default /small \n
        列的宽度是否自撑开
        '''

        
        p = self.get_prop('fit')
        return BoolTypePropInfo(p)

        


    @fit.setter
    def fit(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large / default /small \n
        列的宽度是否自撑开
        '''
        self.set_prop('fit', value)

    @property
    def show_header(self):
        '''
        可选值：large / default /small \n
        是否显示表头
        '''

        
        p = self.get_prop('show-header')
        return BoolTypePropInfo(p)

        


    @show_header.setter
    def show_header(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large / default /small \n
        是否显示表头
        '''
        self.set_prop('show-header', value)

    @property
    def highlight_current_row(self):
        '''
        可选值：large / default /small \n
        是否要高亮当前行
        '''

        
        p = self.get_prop('highlight-current-row')
        return BoolTypePropInfo(p)

        


    @highlight_current_row.setter
    def highlight_current_row(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large / default /small \n
        是否要高亮当前行
        '''
        self.set_prop('highlight-current-row', value)

    @property
    def empty_text(self):
        '''
        可选值：large / default /small \n
        空数据时显示的文本内容， 也可以通过 #empty 设置
        '''

        
        p = self.get_prop('empty-text')
        return StrTypePropInfo(p)

        


    @empty_text.setter
    def empty_text(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：large / default /small \n
        空数据时显示的文本内容， 也可以通过 #empty 设置
        '''
        self.set_prop('empty-text', value)

    @property
    def default_expand_all(self):
        '''
        可选值：large / default /small \n
        是否默认展开所有行，当 Table 包含展开行存在或者为树形表格时有效
        '''

        
        p = self.get_prop('default-expand-all')
        return BoolTypePropInfo(p)

        


    @default_expand_all.setter
    def default_expand_all(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：large / default /small \n
        是否默认展开所有行，当 Table 包含展开行存在或者为树形表格时有效
        '''
        self.set_prop('default-expand-all', value)

    @property
    def default_sort(self):
        '''
        可选值：order: ascending / descending \n
        默认的排序列的 prop 和顺序。 它的 prop 属性指定默认的排序的列，order 指定默认排序的顺序
        '''

        
        return self.get_prop('default-sort')
        


    @default_sort.setter
    def default_sort(self, value: Union[TypePropInfo,Dict]):
        '''
        可选值：order: ascending / descending \n
        默认的排序列的 prop 和顺序。 它的 prop 属性指定默认的排序的列，order 指定默认排序的顺序
        '''
        self.set_prop('default-sort', value)

    @property
    def tooltip_effect(self):
        '''
        可选值：dark / light \n
        tooltip effect 属性
        '''

        
        p = self.get_prop('tooltip-effect')
        return StrTypePropInfo(p)

        


    @tooltip_effect.setter
    def tooltip_effect(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：dark / light \n
        tooltip effect 属性
        '''
        self.set_prop('tooltip-effect', value)

    @property
    def show_summary(self):
        '''
        可选值：dark / light \n
        是否在表尾显示合计行
        '''

        
        p = self.get_prop('show-summary')
        return BoolTypePropInfo(p)

        


    @show_summary.setter
    def show_summary(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：dark / light \n
        是否在表尾显示合计行
        '''
        self.set_prop('show-summary', value)

    @property
    def sum_text(self):
        '''
        可选值：dark / light \n
        合计行第一列的文本
        '''

        
        p = self.get_prop('sum-text')
        return StrTypePropInfo(p)

        


    @sum_text.setter
    def sum_text(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：dark / light \n
        合计行第一列的文本
        '''
        self.set_prop('sum-text', value)

    @property
    def select_on_indeterminate(self):
        '''
        可选值：dark / light \n
        在多选表格中，当仅有部分行被选中时，点击表头的多选框时的行为。 若为 true，则选中所有行；若为 false，则取消选择所有行
        '''

        
        p = self.get_prop('select-on-indeterminate')
        return BoolTypePropInfo(p)

        


    @select_on_indeterminate.setter
    def select_on_indeterminate(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：dark / light \n
        在多选表格中，当仅有部分行被选中时，点击表头的多选框时的行为。 若为 true，则选中所有行；若为 false，则取消选择所有行
        '''
        self.set_prop('select-on-indeterminate', value)

    @property
    def indent(self):
        '''
        可选值：dark / light \n
        展示树形数据时，树节点的缩进
        '''

        
        p = self.get_prop('indent')
        return NumberTypePropInfo(p)

        


    @indent.setter
    def indent(self, value: Union[NumberTypePropInfo,float]):
        '''
        可选值：dark / light \n
        展示树形数据时，树节点的缩进
        '''
        self.set_prop('indent', value)

    @property
    def lazy(self):
        '''
        可选值：dark / light \n
        是否懒加载子节点数据
        '''

        
        p = self.get_prop('lazy')
        return BoolTypePropInfo(p)

        


    @lazy.setter
    def lazy(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：dark / light \n
        是否懒加载子节点数据
        '''
        self.set_prop('lazy', value)

    @property
    def tree_props(self):
        '''
        可选值：dark / light \n
        渲染嵌套数据的配置选项
        '''

        
        return self.get_prop('tree-props')
        


    @tree_props.setter
    def tree_props(self, value: Union[TypePropInfo,Dict]):
        '''
        可选值：dark / light \n
        渲染嵌套数据的配置选项
        '''
        self.set_prop('tree-props', value)

    @property
    def table_layout(self):
        '''
        可选值：fixed / auto \n
        Sets the algorithm used to lay out table cells, rows, and columns
        '''

        
        p = self.get_prop('table-layout')
        return StrTypePropInfo(p)

        


    @table_layout.setter
    def table_layout(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：fixed / auto \n
        Sets the algorithm used to lay out table cells, rows, and columns
        '''
        self.set_prop('table-layout', value)



from typing import TYPE_CHECKING, Any, Dict, List, Union
from pyvisflow.core.props.dataFilterProp import DataFilterPropInfo

from pyvisflow.core.props.fnsMethodProp import FnsMethodPropInfo
from pyvisflow.core.props.methodProp import LambdaMethodPropInfo
from pyvisflow.core.props.typeProp import BoolTypePropInfo, StrTypePropInfo, SubscriptableTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from .components import Component
from pyvisflow.utils.data_gen import get_global_id
from pyvisflow.core.reactive import Reactive
from pyvisflow.core.props.absProp import AbsPropInfo
# if TYPE_CHECKING:
    

class File(Component):
    def __init__(self) -> None:
        super().__init__('file', TComponentType.builtIn)
        self.__info_id=f'data-{get_global_id()}'

    @property
    def columns(self):
        p = self.get_prop('details.columns')
        return SubscriptableTypePropInfo(p)

    @property
    def data(self):
        return FileData(self,self.get_prop('details'))



    def _ex_get_react_data(self) -> Dict[str, Any]:
        data= super()._ex_get_react_data()
        data.update({
            'infoID':self.__info_id,
            'details':{
                'columns':[],
                'data':list(list())
            },
        })
        return data



class FileData(Reactive):

    def __init__(self,file:File,details_binding:AbsPropInfo) -> None:
        super().__init__()
        self.__file=file
        self.details_binding=details_binding
        self.set_prop('details',details_binding)


    def __getitem__(self, column: Union[str, StrTypePropInfo]):
        p = DataFilterPropInfo('row')
        subs = SubscriptableTypePropInfo[str, StrTypePropInfo](p)

        return subs[column]

    def _ex_get_react_data(self) -> Dict[str, Any]:
        data =  super()._ex_get_react_data()
        data.update({
            'details':{
                'columns':[],
                'data':list(list())
            }
        })
        return data

    def filter(self,condition: BoolTypePropInfo):
        lambdaFn = LambdaMethodPropInfo(condition,['row'])
        fns = FnsMethodPropInfo('filter',[self.details_binding,lambdaFn],'dataFns')
        return FileData(self.__file,fns) 

    def orderby(self,columns: Union[List[str],str],asc:Union[List[bool],bool]):
        if isinstance(columns,str):
            columns=[columns]
        if isinstance(asc,bool):
            asc=[asc]    

        if len(asc)>1:
            assert  (len(columns)==len(asc)),'columns len must be eq asc len'
        asc_values = ['asc' if v else 'desc' for v in asc]
        fns = FnsMethodPropInfo('sort',[self.details_binding,columns,asc_values],'dataFns')
        return FileData(self.__file,fns) 

    def groupby(self,keys:List[str]):
        fns = FnsMethodPropInfo('groupby',[self.details_binding,keys],'dataFns')
        return FileData(self.__file,fns) 


from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Generic, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, TypeVar, Union

from pyvisflow.utils.helper import value2code

from .absProp import AbsPropInfo

if TYPE_CHECKING:
    from pyvisflow.core.props.propValuator import Valuator
    from pyvisflow.core.props.typeProp import NumberTypePropInfo, StrTypePropInfo


class IndexPropInfo(AbsPropInfo):
    def __init__(self, valuator: Valuator, parent: Union[AbsPropInfo, None],
                 idx: Union[int, str, StrTypePropInfo, NumberTypePropInfo]
                 ) -> None:
        super().__init__(valuator, parent)
        self.idx = idx

    @staticmethod
    def __to_code(v: Union[str, int, StrTypePropInfo, NumberTypePropInfo]):
        if isinstance(v, (str, int)):
            return value2code(v)
        return v.valuator.cal(v)

    def _ex_gen_expr(self):
        idx = IndexPropInfo.__to_code(self.idx)
        return f'[{idx}]'



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable, List, Optional, Set, Tuple, Dict, TYPE_CHECKING, Generator, Union
from pyvisflow.core.props.methodProp import LambdaMethodPropInfo

from pyvisflow.utils.helper import value2code

from .absProp import AbsPropInfo

from pyvisflow.core.props.propValuator import ConstantsValuator, LambdaValuator
import json


class ValuePropInfo(AbsPropInfo):
    def __init__(self, parent: Union[AbsPropInfo, None],
                 value: Union[float, int, str]) -> None:
        super().__init__(ConstantsValuator(), parent)
        self.value = value

    @staticmethod
    def try2PropInfo(value: Any):
        if (value is None) or isinstance(value, (float, str, int)):
            return ValuePropInfo(None, value)

        if isinstance(value, Dict):
            return JsObjectValuePropInfo(None, value)
        return value

    def _ex_gen_expr(self):
        return value2code(self.value)


class _DictEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, AbsPropInfo):
            m = LambdaMethodPropInfo(o)
            return m.valuator.cal(m)
        return super().default(o)


class DictValuePropInfo(AbsPropInfo):
    def __init__(self, parent: Union[AbsPropInfo, None], value: Dict) -> None:
        super().__init__(ConstantsValuator(), parent)
        self.value = value

    @staticmethod
    def try2PropInfo(value: Any):
        if isinstance(value, (dict)):
            return DictValuePropInfo(None, value)
        return value

    def _ex_gen_expr(self):
        def to_code(key, value):
            if not isinstance(value, (AbsPropInfo)):
                value = value2code(value)
            else:
                m = LambdaMethodPropInfo(value)
                value = m.valuator.cal(m)
            return f'[{value2code(key)},{value}]'

        pairs = [to_code(key, value) for key, value in self.value.items()]
        return f"[{','.join(pairs)}]"


class JsObjectValuePropInfo(AbsPropInfo):
    def __init__(self, parent: Union[AbsPropInfo, None], obj: Dict) -> None:
        super().__init__(ConstantsValuator(), parent)
        self.obj = obj

    @staticmethod
    def try2PropInfo(value: Any):
        if isinstance(value, (dict)):
            return DictValuePropInfo(None, value)
        return value

    def _ex_gen_expr(self):
        def to_code(key, value):
            if not isinstance(value, (AbsPropInfo)):
                value = value2code(value)
            else:
                m = LambdaMethodPropInfo(value)
                value = m.valuator.cal(m)
            return f'{value2code(key)}:{value}'

        pairs = [to_code(key, value) for key, value in self.obj.items()]
        return f"{{{','.join(pairs)}}}"



from .components import *
from .input import *
from .echart import *
from .select import *
from .inputNumber import *
from .dataTable import *
from .containers import *
from .tabs import Tabs
from .text import *
from .icon import Icon
from .file import File


from .app import App

app = App()

__all__ = [
    'box',
    'fns',
    'add_srcipt',
    'to_html',
    'dataTable',
    'input',
    'echart',
    'inputNumber',
    'markdown',
    'plotly',
    'select',
    'cols',
    'tabs',
    'to_html',
    'config2file',
    'text',
    'icon',
    'icon_fromfile',
    'row',
    'file',
]

box = app.box
add_srcipt = app.add_srcipt
dataTable = app.dataTable
input = app.input
echart = app.echart
inputNumber = app.inputNumber
markdown = app.markdown
select = app.select
cols = app.cols
tabs = app.tabs
text = app.text
plotly = app.plotly
fns = app.Fns
icon_fromfile = app.icon_fromfile
icon = app.icon
row=app.row
file=app.file

# v = FnContext()

config2file = app.config2file
to_html = app.to_html



from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyvisflow.core.props.fnsMethodProp import FnsMethodPropInfo
    from pyvisflow.core.props import AbsPropInfo
    from pyvisflow.core.reactive import Reactive
    from pyvisflow.core.props.dataFilterProp import DataFilterPropInfo


class Valuator():
    @abstractmethod
    def cal(self, prop: AbsPropInfo) -> str:
        NotImplementedError()


class LambdaValuator(Valuator):
    def __init__(self) -> None:
        super().__init__()

    def cal(self, prop: AbsPropInfo) -> str:
        return f"()=>{prop.gen_expr()}"


class FnsValuator(Valuator):
    def __init__(self, var: str = 'fns') -> None:
        super().__init__()
        self.var = var

    def cal(self, prop: FnsMethodPropInfo) -> str:
        return f"{self.var}{prop.gen_expr()}"


class DataFilterValuator(Valuator):
    def __init__(self, var: str) -> None:
        super().__init__()
        self.var = var

    def cal(self, prop: DataFilterPropInfo) -> str:
        return f"{prop.gen_expr()}"


class PropValuator(Valuator):
    def __init__(self, reactive: Reactive) -> None:
        super().__init__()
        self.reactive = reactive

    def cal(self, prop: AbsPropInfo):
        return f"getby('{self.reactive._id}',v=> v{prop.gen_expr()})"


class CombineValuator(Valuator):
    def __init__(self) -> None:
        super().__init__()

    def cal(self, prop: AbsPropInfo):
        return prop.gen_expr()


class ConstantsValuator(Valuator):
    def __init__(self) -> None:
        super().__init__()

    def cal(self, prop: AbsPropInfo):
        return prop.gen_expr()



#!/usr/bin/env python
"""The setup script."""

from setuptools import setup, find_packages
import pyvisflow as pvf

with open('README.md', encoding='utf8') as readme_file:
    readme = readme_file.read()

requirements = ['bs4', 'pandas', 'pydantic>=1.9']

test_requirements = [
    'pytest>=3',
]

setup(
    author="pyvisflow",
    author_email='568166495@qq.com',
    python_requires='>=3.6',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="...",
    entry_points={
        # 'console_scripts': [
        #     'test_prj=test_prj.cli:main',
        # ],
    },
    install_requires=requirements,
    license="MIT license",
    # long_description=readme,
    include_package_data=True,
    keywords=['visflow', 'vision', 'flow'],
    name='pyvisflow',
    packages=find_packages(include=['pyvisflow', 'pyvisflow.*']),
    data_files=[('template', [
        'pyvisflow/template/index.html',
        'pyvisflow/template/plotly-2.9.0.min.js'
    ])],
    test_suite='__tests',
    tests_require=test_requirements,
    url='https://gitee.com/carson_add/pyvision-next-example',
    version=pvf.__version__,
    zip_safe=False,
)



from abc import abstractmethod
from typing import Any, Dict, Tuple, Union
from pyvisflow.core.props import NamePropInfo, AbsPropInfo
from pyvisflow.models.TWatchInfo import TValueWatch, TTargetWatchInfo
from pyvisflow.utils.data_gen import get_global_id
import re
from pyvisflow.core.props.propValuator import PropValuator

RE_match_array = re.compile(r'(?P<name>.+)(\[(?P<num>\d+)\])')


class Reactive():
    def __init__(self) -> None:
        self._id = get_global_id()
        self._watch_infos_map: Dict[Tuple, TValueWatch] = {}
        self._free_data: Dict[str, Any] = {}

    def _add_watch_info(self, targetPath: str, value: AbsPropInfo):
        # key = (self._id, targetPath)
        logic = value.valuator.cal(value)
        key = (targetPath, logic)
        if not key in self._watch_infos_map:
            target = TTargetWatchInfo(id=self._id,
                                      path=targetPath,
                                      logic=logic)
            self._watch_infos_map[key] = TValueWatch(target=target)

        # self._watch_infos_map[key].add_trigger(value.to_triggers())

    def get_prop(self, prop: str):
        val = PropValuator(self)
        return NamePropInfo(val, None, prop)

    def set_prop(self, prop: str, value: Union[AbsPropInfo, Any]):
        if isinstance(value, AbsPropInfo):
            self._add_watch_info(prop, value)

            if not prop in self._free_data:
                self._free_data[prop] = None
        else:
            self._free_data[prop] = value
        return self

    def __update_by_path(self, data: Dict, path: str, value: Any):
        paths = path.split('.')
        start = data
        for key in paths[:-1]:

            array = RE_match_array.match(key)
            if array:
                # is array like 'series[0]'
                name, idx = array.group('name'), array.group('num')
                start = start[name][int(idx)]
                continue

            if not key in start:
                start[key] = {}

            start = start[key]

        if paths[-1] in start and (value is
                                   None) and (not start[paths[-1]] is None):
            return

        start.update({paths[-1]: value})

    def _get_WatchInfos(self):
        return list(self._watch_infos_map.values())

    @abstractmethod
    def _ex_get_react_data(self) -> Dict[str, Any]:
        return {}

    def _create_react_data(self) -> Dict[str, Any]:
        data = self._ex_get_react_data()

        for prop, value in self._free_data.items():
            self.__update_by_path(data, prop, value)
        return data



from typing import Union,Dict,List
from pyvisflow.core.components.containers import Container
from pyvisflow.core.props import TypePropInfo,StrTypePropInfo,SubscriptableTypePropInfo, NumberTypePropInfo, BoolTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.core.props import *


class _Tabs(Container):
    '''
    分隔内容上有关联但属于不同类别的数据集合
    '''

    def __init__(self) -> None:
        super().__init__('tabs', TComponentType.builtIn)

    def _ex_get_react_data(self):
        return super()._ex_get_react_data()


    @property
    def names(self):
        '''
        可选值：nan \n
        选项卡名字
        '''

        
        p = self.get_prop('names')
        return SubscriptableTypePropInfo[int,StrTypePropInfo](p)

        


    @names.setter
    def names(self, value: Union[SubscriptableTypePropInfo[int,StrTypePropInfo],List[str]]):
        '''
        可选值：nan \n
        选项卡名字
        '''
        self.set_prop('names', value)

    @property
    def activeName(self):
        '''
        可选值：nan \n
        当前激活的选项卡名字
        '''

        
        p = self.get_prop('activeName')
        return StrTypePropInfo(p)

        


    @activeName.setter
    def activeName(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：nan \n
        当前激活的选项卡名字
        '''
        self.set_prop('activeName', value)

    @property
    def value(self):
        '''
        可选值：nan \n
        绑定值，选中选项卡的 name
        '''

        
        p = self.get_prop('value')
        return StrTypePropInfo(p)

        


    @value.setter
    def value(self, value: Union[StrTypePropInfo,Union[str,float]]):
        '''
        可选值：nan \n
        绑定值，选中选项卡的 name
        '''
        self.set_prop('value', value)

    @property
    def type(self):
        '''
        可选值：card/border-card \n
        风格类型
        '''

        
        p = self.get_prop('type')
        return StrTypePropInfo(p)

        


    @type.setter
    def type(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：card/border-card \n
        风格类型
        '''
        self.set_prop('type', value)

    @property
    def closable(self):
        '''
        可选值：card/border-card \n
        标签是否可关闭
        '''

        
        p = self.get_prop('closable')
        return BoolTypePropInfo(p)

        


    @closable.setter
    def closable(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：card/border-card \n
        标签是否可关闭
        '''
        self.set_prop('closable', value)

    @property
    def addable(self):
        '''
        可选值：card/border-card \n
        标签是否可增加
        '''

        
        p = self.get_prop('addable')
        return BoolTypePropInfo(p)

        


    @addable.setter
    def addable(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：card/border-card \n
        标签是否可增加
        '''
        self.set_prop('addable', value)

    @property
    def editable(self):
        '''
        可选值：card/border-card \n
        标签是否同时可增加和关闭
        '''

        
        p = self.get_prop('editable')
        return BoolTypePropInfo(p)

        


    @editable.setter
    def editable(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：card/border-card \n
        标签是否同时可增加和关闭
        '''
        self.set_prop('editable', value)

    @property
    def tab_position(self):
        '''
        可选值：top/right/bottom/left \n
        选项卡所在位置
        '''

        
        p = self.get_prop('tab-position')
        return StrTypePropInfo(p)

        


    @tab_position.setter
    def tab_position(self, value: Union[StrTypePropInfo,str]):
        '''
        可选值：top/right/bottom/left \n
        选项卡所在位置
        '''
        self.set_prop('tab-position', value)

    @property
    def stretch(self):
        '''
        可选值：top/right/bottom/left \n
        标签的宽度是否自撑开
        '''

        
        p = self.get_prop('stretch')
        return BoolTypePropInfo(p)

        


    @stretch.setter
    def stretch(self, value: Union[BoolTypePropInfo,bool]):
        '''
        可选值：top/right/bottom/left \n
        标签的宽度是否自撑开
        '''
        self.set_prop('stretch', value)



from typing import Dict, List, Optional, Union
from pyvisflow.core.components.components import Component
from pyvisflow.core.dataManager import DataFrameManager, TDataFrameInfo
from pyvisflow.core.props import StrTypePropInfo, SubscriptableTypePropInfo, BoolTypePropInfo
from pyvisflow.core.props.dataFilterProp import DataFilterPropInfo
from pyvisflow.core.props.nameProp import NamePropInfo
from pyvisflow.models.TWatchInfo import TTableFilterWatch, TTableFilterTarget
from pyvisflow.utils.helper import df2object_dict
from .auto_create._dataTable import _DataTable
import pandas as pd


class DataTable(_DataTable):
    def __init__(self, data_info: TDataFrameInfo) -> None:
        super().__init__()

        self._data_id = data_info
        self.current_page = 1
        self._filter_watch_infos: List[TTableFilterWatch] = []

    def __getitem__(self, column: Union[str, StrTypePropInfo]):
        p = DataFilterPropInfo('row')
        subs = SubscriptableTypePropInfo[str, StrTypePropInfo](p)

        return subs[column]

    def query(self, condition: BoolTypePropInfo):
        logic = condition.valuator.cal(condition)
        target = TTableFilterTarget(id=self._id, logic=logic)
        filter = TTableFilterWatch(target=target)
        self._filter_watch_infos.append(filter)

    @property
    def rowClick(self):
        p = self.get_prop('rowClick')
        return SubscriptableTypePropInfo[str, StrTypePropInfo](p)

    @property
    def rowHover(self):
        p = self.get_prop('rowHover')
        return SubscriptableTypePropInfo[str, StrTypePropInfo](p)

    def _ex_get_react_data(self):
        data = super()._ex_get_react_data()
        data.update({
            'dataInfo': self._data_id,
            'rowClick': {},
            'rowHover': {},
        })
        return data



from pyvisflow.core.components.components import Component
from pyvisflow.core.props import TypePropInfo
from pyvisflow.models.TComponent import TComponentType


class MarkdownValue(Component):
    def __init__(self) -> None:
        super().__init__('markdown-value', TComponentType.builtIn)

    def set_value(self, prop: TypePropInfo):
        self.set_prop('value', prop)



from typing import Union
from pyvisflow.core.props.typeProp import StrTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from .components import Component


class Text(Component):
    def __init__(self, value:str) -> None:
        super().__init__('text', TComponentType.builtIn)
        self.set_prop('value',value)

    @property
    def value(self):
        '''
        文字内容
        '''

        p = self.get_prop('value')
        return StrTypePropInfo(p)

    @value.setter
    def value(self, value: Union[StrTypePropInfo, str]):
        '''
        文字内容
        '''
        self.set_prop('value', value)


from typing import Union, Dict, List
from pyvisflow.core.components.components import Component
from pyvisflow.core.props import TypePropInfo, StrTypePropInfo, SubscriptableTypePropInfo, NumberTypePropInfo, BoolTypePropInfo
from pyvisflow.models.TComponent import TComponentType
from pyvisflow.core.props import *


class _Select(Component):
    '''
    Select 选择器
当选项过多时，使用下拉菜单展示并选择内容
    '''
    def __init__(self) -> None:
        super().__init__('select', TComponentType.builtIn)

    def _ex_get_react_data(self):
        return super()._ex_get_react_data()

    @property
    def multiple(self):
        '''
        可选值：— \n
        是否多选
        '''

        p = self.get_prop('multiple')
        return BoolTypePropInfo(p)

    @multiple.setter
    def multiple(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：— \n
        是否多选
        '''
        self.set_prop('multiple', value)

    @property
    def disabled(self):
        '''
        可选值：— \n
        是否禁用
        '''

        p = self.get_prop('disabled')
        return BoolTypePropInfo(p)

    @disabled.setter
    def disabled(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：— \n
        是否禁用
        '''
        self.set_prop('disabled', value)

    @property
    def value_key(self):
        '''
        可选值：— \n
        作为 value 唯一标识的键名，绑定值为对象类型时必填
        '''

        p = self.get_prop('value-key')
        return StrTypePropInfo(p)

    @value_key.setter
    def value_key(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：— \n
        作为 value 唯一标识的键名，绑定值为对象类型时必填
        '''
        self.set_prop('value-key', value)

    @property
    def size(self):
        '''
        可选值：large/default/small \n
        输入框尺寸
        '''

        p = self.get_prop('size')
        return StrTypePropInfo(p)

    @size.setter
    def size(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：large/default/small \n
        输入框尺寸
        '''
        self.set_prop('size', value)

    @property
    def clearable(self):
        '''
        可选值：large/default/small \n
        是否可以清空选项
        '''

        p = self.get_prop('clearable')
        return BoolTypePropInfo(p)

    @clearable.setter
    def clearable(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：large/default/small \n
        是否可以清空选项
        '''
        self.set_prop('clearable', value)

    @property
    def collapse_tags(self):
        '''
        可选值：large/default/small \n
        多选时是否将选中值按文字的形式展示
        '''

        p = self.get_prop('collapse-tags')
        return BoolTypePropInfo(p)

    @collapse_tags.setter
    def collapse_tags(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：large/default/small \n
        多选时是否将选中值按文字的形式展示
        '''
        self.set_prop('collapse-tags', value)

    @property
    def multiple_limit(self):
        '''
        可选值：large/default/small \n
        多选时用户最多可以选择的项目数， 为 0 则不限制
        '''

        p = self.get_prop('multiple-limit')
        return NumberTypePropInfo(p)

    @multiple_limit.setter
    def multiple_limit(self, value: Union[NumberTypePropInfo, int]):
        '''
        可选值：large/default/small \n
        多选时用户最多可以选择的项目数， 为 0 则不限制
        '''
        self.set_prop('multiple-limit', value)

    @property
    def name(self):
        '''
        可选值：large/default/small \n
        多选框的输入框的原生 name 属性
        '''

        p = self.get_prop('name')
        return StrTypePropInfo(p)

    @name.setter
    def name(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：large/default/small \n
        多选框的输入框的原生 name 属性
        '''
        self.set_prop('name', value)

    @property
    def effect(self):
        '''
        可选值：string \n
        Tooltip theme, built-in theme: dark / light
        '''

        p = self.get_prop('effect')
        return StrTypePropInfo(p)

    @effect.setter
    def effect(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        Tooltip theme, built-in theme: dark / light
        '''
        self.set_prop('effect', value)

    @property
    def autocomplete(self):
        '''
        可选值：string \n
        ‎选择输入的自动完成属性‎
        '''

        p = self.get_prop('autocomplete')
        return StrTypePropInfo(p)

    @autocomplete.setter
    def autocomplete(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        ‎选择输入的自动完成属性‎
        '''
        self.set_prop('autocomplete', value)

    @property
    def placeholder(self):
        '''
        可选值：string \n
        ‎占 位 符‎
        '''

        p = self.get_prop('placeholder')
        return StrTypePropInfo(p)

    @placeholder.setter
    def placeholder(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        ‎占 位 符‎
        '''
        self.set_prop('placeholder', value)

    @property
    def filterable(self):
        '''
        可选值：string \n
        ‎选择是否可过滤‎
        '''

        p = self.get_prop('filterable')
        return BoolTypePropInfo(p)

    @filterable.setter
    def filterable(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：string \n
        ‎选择是否可过滤‎
        '''
        self.set_prop('filterable', value)

    @property
    def allow_create(self):
        '''
        可选值：string \n
        ‎是否允许创建新项目。要使用此，必须为 true‎filterable
        '''

        p = self.get_prop('allow-create')
        return BoolTypePropInfo(p)

    @allow_create.setter
    def allow_create(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：string \n
        ‎是否允许创建新项目。要使用此，必须为 true‎filterable
        '''
        self.set_prop('allow-create', value)

    @property
    def remote(self):
        '''
        可选值：string \n
        ‎是否从服务器加载选项‎
        '''

        p = self.get_prop('remote')
        return BoolTypePropInfo(p)

    @remote.setter
    def remote(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：string \n
        ‎是否从服务器加载选项‎
        '''
        self.set_prop('remote', value)

    @property
    def loading(self):
        '''
        可选值：string \n
        ‎选择是否正在从服务器加载数据‎
        '''

        p = self.get_prop('loading')
        return BoolTypePropInfo(p)

    @loading.setter
    def loading(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：string \n
        ‎选择是否正在从服务器加载数据‎
        '''
        self.set_prop('loading', value)

    @property
    def loading_text(self):
        '''
        可选值：string \n
        ‎从服务器加载数据时显示的文本‎
        '''

        p = self.get_prop('loading-text')
        return StrTypePropInfo(p)

    @loading_text.setter
    def loading_text(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        ‎从服务器加载数据时显示的文本‎
        '''
        self.set_prop('loading-text', value)

    @property
    def no_match_text(self):
        '''
        可选值：string \n
        ‎当没有数据与过滤查询匹配时显示的文本，也可以使用 slot‎empty
        '''

        p = self.get_prop('no-match-text')
        return StrTypePropInfo(p)

    @no_match_text.setter
    def no_match_text(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        ‎当没有数据与过滤查询匹配时显示的文本，也可以使用 slot‎empty
        '''
        self.set_prop('no-match-text', value)

    @property
    def no_data_text(self):
        '''
        可选值：string \n
        ‎显示文本时没有选项，也可以使用插槽‎empty
        '''

        p = self.get_prop('no-data-text')
        return StrTypePropInfo(p)

    @no_data_text.setter
    def no_data_text(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        ‎显示文本时没有选项，也可以使用插槽‎empty
        '''
        self.set_prop('no-data-text', value)

    @property
    def popper_class(self):
        '''
        可选值：string \n
        ‎"选择"下拉列表的自定义类名‎
        '''

        p = self.get_prop('popper-class')
        return StrTypePropInfo(p)

    @popper_class.setter
    def popper_class(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：string \n
        ‎"选择"下拉列表的自定义类名‎
        '''
        self.set_prop('popper-class', value)

    @property
    def reserve_keyword(self):
        '''
        可选值：string \n
        ‎何时 为 true，是否在选择选项后保留当前关键字‎multiplefilter
        '''

        p = self.get_prop('reserve-keyword')
        return BoolTypePropInfo(p)

    @reserve_keyword.setter
    def reserve_keyword(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：string \n
        ‎何时 为 true，是否在选择选项后保留当前关键字‎multiplefilter
        '''
        self.set_prop('reserve-keyword', value)

    @property
    def default_first_option(self):
        '''
        可选值：string \n
        ‎在输入键上选择第一个匹配选项。与 或 一起使用‎filterableremote
        '''

        p = self.get_prop('default-first-option')
        return BoolTypePropInfo(p)

    @default_first_option.setter
    def default_first_option(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：string \n
        ‎在输入键上选择第一个匹配选项。与 或 一起使用‎filterableremote
        '''
        self.set_prop('default-first-option', value)

    @property
    def teleported(self):
        '''
        可选值：true / false \n
        ‎选择下拉列表是否传送到正文‎
        '''

        p = self.get_prop('teleported')
        return BoolTypePropInfo(p)

    @teleported.setter
    def teleported(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：true / false \n
        ‎选择下拉列表是否传送到正文‎
        '''
        self.set_prop('teleported', value)

    @property
    def persistent(self):
        '''
        可选值：true / false \n
        ‎当选择下拉列表处于非活动状态并且是 时，选择下拉列表将被销毁‎persistentfalse
        '''

        p = self.get_prop('persistent')
        return BoolTypePropInfo(p)

    @persistent.setter
    def persistent(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：true / false \n
        ‎当选择下拉列表处于非活动状态并且是 时，选择下拉列表将被销毁‎persistentfalse
        '''
        self.set_prop('persistent', value)

    @property
    def automatic_dropdown(self):
        '''
        可选值：true / false \n
        ‎对于不可过滤的 Select，此属性决定在输入聚焦时是否弹出选项菜单‎
        '''

        p = self.get_prop('automatic-dropdown')
        return BoolTypePropInfo(p)

    @automatic_dropdown.setter
    def automatic_dropdown(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：true / false \n
        ‎对于不可过滤的 Select，此属性决定在输入聚焦时是否弹出选项菜单‎
        '''
        self.set_prop('automatic-dropdown', value)

    @property
    def fit_input_width(self):
        '''
        可选值：true / false \n
        ‎下拉列表的宽度是否与输入相同‎
        '''

        p = self.get_prop('fit-input-width')
        return BoolTypePropInfo(p)

    @fit_input_width.setter
    def fit_input_width(self, value: Union[BoolTypePropInfo, bool]):
        '''
        可选值：true / false \n
        ‎下拉列表的宽度是否与输入相同‎
        '''
        self.set_prop('fit-input-width', value)

    @property
    def tag_type(self):
        '''
        可选值：success/info/warning/danger \n
        ‎标记类型‎
        '''

        p = self.get_prop('tag-type')
        return StrTypePropInfo(p)

    @tag_type.setter
    def tag_type(self, value: Union[StrTypePropInfo, str]):
        '''
        可选值：success/info/warning/danger \n
        ‎标记类型‎
        '''
        self.set_prop('tag-type', value)



from abc import abstractmethod
from typing import Any, Dict, List, Tuple, Union
from pyvisflow.core.props.typeProp import StrTypePropInfo
from pyvisflow.models.TComponent import TComponent, TComponentType
import re

from pyvisflow.models.TWatchInfo import TChartFilterWatch
from pyvisflow.core.reactive import Reactive

RE_match_array = re.compile(r'(?P<name>.+)(\[(?P<num>\d+)\])')


class Component(Reactive):
    def __init__(self, tag: str, type: TComponentType) -> None:
        super().__init__()
        self._tag = tag
        self._type = type
        self._attrs = {}
        self._styles = {}
        self._children: List[Component] = []
        self._stylesProp = StyleProp(self)

    @property
    def styles(self):
        return self._stylesProp

    def update_attrs(self, **kws):
        self._attrs.update(kws)
        return self

    def update_styles(self, **kws):

        for key, value in kws.items():
            self.set_prop(f'styles.{key}', value)

        # self._styles.update(kws)
        return self

    def _ex_get_react_data(self) -> Dict[str, Any]:
        if self._type == TComponentType.builtIn:
            data = {
                'attrs': self._attrs,
                'styles': self._styles,
            }

            return data

        return {}

    def to_model(self) -> TComponent:
        children = [c.to_model() for c in self._children]
        return TComponent(id=self._id,
                          type=self._type,
                          tag=self._tag,
                          attrs=self._attrs,
                          styles=self._styles,
                          children=children)


class StyleProp():
    def __init__(self, cp: Component) -> None:
        self.__cp = cp

    def set(self, name: str, value: Union[StrTypePropInfo, str]):
        self.__cp.set_prop(f'styles.{name}', value)
        return self

    @property
    def width(self):
        '''
        width 属性用于设置元素的宽度
        /* <length> values */
        width: 300px;
        width: 25em;

        /* <percentage> value */
        width: 75%;

        /* Keyword values */
        width: max-content;
        width: min-content;
        width: fit-content(20em);
        width: auto;

        /* Global values */
        width: inherit;
        width: initial;
        width: unset;
        '''
        p = self.__cp.get_prop('styles.width')
        return StrTypePropInfo(p)

    def make_center(self):
        self.set('display', 'flex').set('justify-content',
                                        'center').set('align-items', 'center')
        return self

    def set_width(self, value: Union[StrTypePropInfo, str]):
        self.__cp.set_prop('styles.width', value)
        return self

    def set_height(self, value: Union[StrTypePropInfo, str]):
        self.__cp.set_prop('styles.height', value)
        return self

    def set_fixed(
            self,
            top: Union[StrTypePropInfo, str] = 'unset',
            right: Union[StrTypePropInfo, str] = 'unset',
            bottom: Union[StrTypePropInfo, str] = 'unset',
            left: Union[StrTypePropInfo, str] = 'unset',
    ):
        '''
        元素的位置在屏幕滚动时不会改变。打印时，元素会出现在的每页的固定位置
        '''
        self.__cp.set_prop('styles.position', 'fixed')
        self.__cp.set_prop('styles.top', top)
        self.__cp.set_prop('styles.right', right)
        self.__cp.set_prop('styles.bottom', bottom)
        self.__cp.set_prop('styles.left', left)
        return self

    def set_absolute(
            self,
            top: Union[StrTypePropInfo, str] = '0',
            right: Union[StrTypePropInfo, str] = 'unset',
            bottom: Union[StrTypePropInfo, str] = 'unset',
            left: Union[StrTypePropInfo, str] = '0',
    ):
        '''
        元素会被移出正常文档流，并不为元素预留空间，通过指定元素相对于最近的非 static 定位祖先元素的偏移，来确定元素位置。绝对定位的元素可以设置外边距（margins），且不会与其他边距合并
        '''
        self.__cp.set_prop('styles.position', 'absolute')
        self.__cp.set_prop('styles.top', top)
        self.__cp.set_prop('styles.right', right)
        self.__cp.set_prop('styles.bottom', bottom)
        self.__cp.set_prop('styles.left', left)

        self.set_height('100%')
        return self

    def set_absolute_center(self):
        '''
        justify-content: center;
        align-items: center;
        left: 50%;
        top: 50%;
        transform: translateX(-50%) translateY(-50%);
        '''
        self.set('display', 'flex').set('justify-content', 'center').set(
            'align-items',
            'center').set('transform',
                          'translateX(-50%) translateY(-50%)').set(
                              'left', '50%').set('top', '50%')
        return self

    def set_absolute_left(self):
        '''
        justify-content: center;
        align-items: center;
        left: 0;
        top: 50%;
        transform: translateY(-50%);
        '''
        self.set('display', 'flex').set('justify-content', 'center').set(
            'align-items', 'center').set('transform',
                                         'translateX(0) translateY(-50%)').set(
                                             'top', '50%').set('left', '0')
        return self

    def set_border(self,
                   value: Union[StrTypePropInfo, str] = '1px solid black'):
        '''
        border属性是一个用于设置各种单独的边界属性的简写属性。border可以用于设置一个或多个以下属性的值： border-width, border-style, border-color。
        '''
        self.__cp.set_prop('styles.border', value)
        return self

    def set_margin(self, value: Union[StrTypePropInfo, str] = '1rem'):
        '''
        margin 属性为给定元素设置所有四个（上下左右）方向的外边距属性。也就是 margin-top，margin-right，margin-bottom，和 margin-left 四个外边距属性设置的简写
        '''
        self.__cp.set_prop('styles.margin', value)
        return self

    def set_background(self, value: Union[StrTypePropInfo, str] = 'unset'):
        '''
        background 是一种 CSS 简写属性，用于一次性集中定义各种背景属性，包括 color, image, origin 与 size, repeat 方式等等。
        '''
        self.__cp.set_prop('styles.background', value)
        return self

    def set_border_radius(self, value: Union[StrTypePropInfo, str] = '2rem'):
        '''
        border-radius 允许你设置元素的外边框圆角。当使用一个半径时确定一个圆形，当使用两个半径时确定一个椭圆。这个(椭)圆与边框的交集形成圆角效果
        '''
        self.__cp.set_prop('styles.border-radius', value)
        return self

    def set_transition(self, value: Union[StrTypePropInfo, str] = '0.6s'):
        '''
        transition CSS 属性是 transition-property，transition-duration，transition-timing-function 和 transition-delay 的一个简写属性。        '''
        self.__cp.set_prop('styles.transition', value)
        return self